import os
from .operation import Operation
from ..errors import ValidationError


def _PASS(x): return "_".join(x.split("/"))


class TravisPrefixGeneration(Operation, register="travisprefixgen"):
    """
        Generates a deterministic key prefix from Travis metadata
    """
    HIDDEN = True
    SILENT = True

    ENVMAP = {
        "TRAVIS_COMMIT": ("commit", _PASS),
        "TRAVIS_REPO_SLUG": ("repo", _PASS),
        "TRAVIS_BRANCH": ("branch", _PASS),
        "TRAVIS_PULL_REQUEST": ("push", lambda x: x == "false")
    }

    @staticmethod
    def get_pull_request_sha():
        return os.environ.get("TRAVIS_PULL_REQUEST_SHA", None)

    @classmethod
    def get_pull_request(cls):
        retval = os.environ.get("TRAVIS_PULL_REQUEST", "false")
        if retval == "false":
            return None
        return retval

    def prevalidate(self, args):
        valid = True

        for envvar in self.ENVMAP.keys():
            if envvar not in os.environ:
                valid = False
                break

        if not valid or (self.get_pull_request()
                         and not self.get_pull_request_sha()):
            raise ValidationError(
                "Expected TravisCI variables missing!",
                False
            )

    def prepare_args(self, args):
        return {arg: xform(os.environ[envvar])
                for envvar, (arg, xform) in self.ENVMAP.items()}

    def run(self, commit: str, repo: str, branch: str, push: bool):
        prefix = f"travis/{repo}/{branch}"

        if push:
            prefix += f"/push/{commit}"
        else:
            prefix += f"/pull/{commit}/{self.get_pull_request()}/{self.get_pull_request_sha()}"  # noqa

        self.console.stream.write(prefix)



import os
import json
from typing import TextIO, Iterable


class Console:
    def __init__(self, stream: TextIO, linesep=os.linesep):
        self.stream = stream
        self.linesep = linesep

    def print(self, text: str):
        self.stream.write(f"{text} {self.linesep}")

    def print_title(self, title):
        self.newln()
        self.print(f"{title}:")

    def print_iter(self, title: str, items: Iterable):
        self.print_title(title)
        self.newln()
        for item in items:
            self.print(f" - {item}")
        self.newln()

    def print_json(self, title: str, obj):
        self.print_title(title)
        self.newln()
        json.dump(obj, self.stream, indent=4)
        self.newln()

    def newln(self):
        self.print("")



import hyron
from .operation import Operation
from ..errors import ValidationError


class RenderArtifact(Operation, register="render"):
    """
        Render an artifact to console.
        Usage: hyrontools render [rulebookfile] [artifact]
    """

    ARGC = 2

    def prepare_args(self, args: tuple):
        loader = hyron.rulebooks.RulebookLoader()
        return {
            "rulebook": loader.load(args[0]),
            "artifact": args[1],
        }

    def postvalidate(self, rulebook: hyron.rulebooks.Rulebook, artifact: str):
        self.assert_artifacts(rulebook)
        if artifact not in rulebook.artifacts:
            raise ValidationError(
                f"Rulebook '{rulebook.title}' doesn't reference an artifact named '{artifact}'",  # noqa
                False
            )

    def run(self, rulebook: hyron.rulebooks.Rulebook, artifact: str):
        self.console.print(f"Rendering '{artifact}' artifact...")

        artifact = rulebook.build_artifact(artifact)

        self.console.print("Render complete.")

        self.console.print_json("Artifact metadata", artifact.meta)

        self.console.print_iter("File list", artifact.files.keys())

        for name, artifact_file in artifact.files.items():
            encoding = artifact_file.encoding
            self.console.print_title(f"File '{name}'")
            self.console.newln()
            self.console.print(artifact_file.get_contents().decode(encoding))
            self.console.newln()



class CliSyntaxError(RuntimeError):
    """
        Represents an error in a CLI operation syntax.
        Will trigger CLI to display usage info.
    """

    def __init__(self, invocation: str, msg: str):
        self.invocation = invocation
        self.msg = msg

    def __str__(self):
        return f"Syntax error for '{self.invocation}' operation: {self.msg}"



import os
from dataclasses import field


def load_file(filename, handler):
    with open(filename, 'r') as infile:
        return handler(infile)


def get_base_dir(filename):
    return os.path.dirname(os.path.realpath(filename))


def get_pkg_dir():
    return get_base_dir(__file__)


def get_pkg_filename(*args):
    return os.path.join(get_pkg_dir(), *args)


def get_asset_filename(name, ext="txt"):
    return get_pkg_filename("assets", f"{name}.{ext}")


def load_text_asset(name):
    return load_file(get_asset_filename(name), lambda x: x.read())


def default(factory):
    return field(default_factory=factory)



import boto3
import hyron
import tempfile
from botocore.exceptions import ClientError
from .operation import Operation
from ..errors import ValidationError


class S3PublishArtifacts(Operation, register="s3publish"):
    """
        Package all artifacts in a rulebook and publish them to Amazon S3.
        Filename format will be prefix/artifactname.tar.gz
        Usage: hyrontools publish [rulebookfile] [bucket] [prefix]
    """

    ARGC = 3

    @staticmethod
    def get_key(prefix, filename):
        return f"{prefix}/{filename}"

    def get_s3_resource(self):
        aws = boto3.Session()
        return aws.resource("s3")

    def prepare_args(self, args: tuple):
        loader = hyron.rulebooks.RulebookLoader()
        resource = self.get_s3_resource()

        return {
            "rulebook": loader.load(args[0]),
            "bucket": resource.Bucket(args[1]),
            "prefix": args[2]
        }

    def postvalidate(
            self,
            rulebook: hyron.rulebooks.Rulebook,
            bucket,
            prefix: str):
        self.assert_artifacts(rulebook)
        try:
            bucket.put_object(
                Key=self.get_key(
                    prefix,
                    ".test"),
                Body=b"test",
                ACL="bucket-owner-full-control")
        except ClientError:
            raise ValidationError(
                f"Unable to write to {bucket.name}, check your credentials.",
                False
            )

    def run(self, rulebook: hyron.rulebooks.Rulebook, bucket, prefix: str):
        self.console.print("Rendering all artifacts...")
        artifacts = rulebook.build_all()
        count = len(artifacts)
        self.console.print(f"Rendering complete: {count} artifacts created.")
        self.console.newln()
        self.console.print("Starting package and publish sequence")
        with tempfile.TemporaryDirectory() as tmpdir:
            for job, (name, artifact) in enumerate(artifacts.items()):
                self.console.print(
                    f"Beginning processing for '{name}' artifact.")
                key = self.get_key(prefix, f"{name}.tar.gz")
                local_package = artifact.to_archive(tmpdir)
                self.console.print(f"Package built: {local_package}")
                self.console.print(
                    f"Uploading package to s3://{bucket.name}/{key}...")
                with open(local_package, "rb") as infile:
                    bucket.put_object(
                        Key=key, Body=infile, ACL="bucket-owner-full-control")
                self.console.print(
                    f"Upload complete - {job+1}/{count} packages have been published."  # noqa
                )
        self.console.print("All packages have been published")



import sys

from .console import Console
from . import operations, helpers, errors


class Executor:
    def __init__(self, console: Console):
        self.console = console
        self.ops = operations.Operation.registry.enum()
        self.script = "hyrontools"

    def print_welcome(self):
        self.print_cli_asset("header")
        self.console.print(f" - Python runtime: {sys.executable}")
        self.console.print(f" - Script name: {self.script}")
        self.print_cli_asset("breakline")

    def get_operation(self, name, check=True):
        if check and name not in self.ops:
            raise errors.CliExecutionError(
                f"Operation '{name}'' is undefined.")
        return operations.Operation.get(name, name, self.console)

    def dispatch(self, *args):
        op = self.get_operation(args[0])
        if not op.SILENT:
            self.print_welcome()
        result = op.parse_args(*args[1:])
        if not result.valid:
            result.raise_error()
        op.run(**result.args)

    def print_error(self, exception):
        self.console.newln()
        self.console.print(str(exception))
        self.console.newln()

    def print_cli_asset(self, asset):
        self.console.print(helpers.load_text_asset(f"cli_{asset}"))

    def print_usage(self):
        operations = map(lambda x: self.get_operation(x, False), self.ops)
        self.console.print_iter(
            "Usage", [op.usage for op in operations if not op.HIDDEN]
        )

    def entrypoint(self, *args):
        self.script = args[0]
        args = args[1:]

        try:
            if not args:
                raise errors.CliExecutionError(silent=True)
            self.dispatch(*args)
        except errors.CliExecutionError as e:
            if not e.silent:
                self.print_error(e)
            if e.showusage:
                self.print_usage()
            return 1
        except errors.CliRuntimeError as e:
            self.print_error(e)
            return 2
        except errors.CliSyntaxError as e:
            self.print_error(e)
            self.print_usage()
            return 3

        return 0



from .runtime import CliRuntimeError
from .syntax import CliSyntaxError
from .validation import ValidationError
from .exec import CliExecutionError

__all__ = [
    "CliRuntimeError",
    "CliSyntaxError",
    "ValidationError",
    "CliExecutionError"
]



from abc import abstractmethod
from textwrap import dedent, indent
from plugable import Plugable

from ..console import Console
from ..argument_parse_result import ArgumentParseResult
from ..errors import ValidationError


class Operation(Plugable):
    """
        This abstract class represents a CLI operation.
        CLI operations use their docstrings as help text,
        and are invoked using their class registry names.
    """

    ARGC = 0  # If you have an set length of expected args
    ARGV = ()  # Tuple of arg names for set length ops
    HIDDEN = False  # Hidden ops don't show in usage
    SILENT = False  # Silent ops mute executor output

    def __init__(self, invocation: str, console: Console):
        self.invocation = invocation
        self.console = console

    @abstractmethod
    def run(self, **kwargs):
        raise NotImplementedError()

    def postvalidate(self, **kwargs):
        pass

    def prepare_args(self, args):
        return dict(zip(args, self.ARGV))

    def prevalidate(self, args):
        if len(args) != self.ARGC:
            raise ValidationError(
                f"'{self.invocation}' takes {self.ARGC} arguments")

    def parse_args(self, *args) -> ArgumentParseResult:
        result = ArgumentParseResult(self.invocation)

        try:
            self.prevalidate(args)
            kwargs = self.prepare_args(args)
            self.postvalidate(**kwargs)
            result.args = kwargs
            result.valid = True
        except ValidationError as e:
            result.error = e
            result.valid = False

        return result

    @property
    def usage(self):
        linesep = self.console.linesep
        usage = indent(dedent(self.__doc__).strip(), "    | ")

        return f"{self.invocation}:{linesep}{usage}{linesep}"

    @staticmethod
    def assert_artifacts(rulebook):
        if not rulebook.artifacts:
            raise ValidationError(
                f"Rulebook '{rulebook.title}' contain any artifacts",
                False
            )



from hyron.renderers.renderer import Renderer
from .operation import Operation


class ListRenderers(Operation, register="renderers"):
    """
        List the available renderers.
    """

    def run(self, *args):
        self.console.print_iter(
            "Available renderers",
            Renderer.registry.enum())



from .operation import Operation
from .renderers import ListRenderers
from .render import RenderArtifact
from .s3publish import S3PublishArtifacts
from .travisprefixgen import TravisPrefixGeneration

__all__ = [
    "Operation",
    "ListRenderers",
    "RenderArtifact",
    "S3PublishArtifacts",
    "TravisPrefixGeneration"
]



class ValidationError(RuntimeError):
    """
        Represents a validation failure during the argumaent parse phase
    """

    def __init__(self, msg: str, syntax=True):
        self.msg = msg
        self.syntax = syntax



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hyrontools",
    version="1.0.0",
    author="Jacob Neil Taylor",
    author_email="me@jacobtaylor.id.au",
    description="Hyron CLI Tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jacobneiltaylor/hyrontools",
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    package_data={
        "hyrontools": [
            "assets/*"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",  # noqa
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'hyrontools = hyrontools.__main__:main',
        ],
    },
    python_requires='>=3.7',
    install_requires=[
        "hyron",
        "boto3"
    ],
)



from dataclasses import dataclass
from .helpers import default
from .errors import CliSyntaxError, CliRuntimeError, ValidationError


@dataclass
class ArgumentParseResult:
    invocation: str
    valid: bool = False
    args: dict = default(dict)
    error: ValidationError = None

    def raise_error(self):
        if self.error.syntax:
            raise CliSyntaxError(self.invocation, self.error.msg)
        raise CliRuntimeError(self.invocation, self.error.msg)



class CliExecutionError(RuntimeError):
    """
        Represents an error encountered in the executor.
    """

    def __init__(self, msg: str = "", silent=False, showusage=True):
        self.msg = msg
        self.silent = silent
        self.showusage = showusage

    def __str__(self):
        return self.msg
